#include "OpenGLScene.h"

#include <QVector3D>

#include "Shape.h"
#include "Arrow.h"
#include "FruitPole.h"
#include "Triangle.h"
#include "Cube.h"
#include "Quad.h"
#include "String3D.h"
#include "SphericalCollider.h"
#include "CollisionChecker.h"
#include "DialogHUD.h"
#include "DisplayHUD.h"

#include "ScoreUI.h"
#include "Stopwatch.h"

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

struct OpenGLSceneData
{
	Arrow arrow;
	FruitPole fruitPole;
	String3D target;
	Sphere fruit;
	SphericalCollider arrowCollider;
	Quad background;

	GLdouble targetRotAngle;
	GLdouble projectileDisplaceX;
	GLdouble projectileDisplaceY;
	GLfloat fruitFall;
	int toggleTargetIncrDecrFlag;
	int toggleTargetHit;

	QTimer* targetTimer;
	QTimer* projectileTimer;

	QPointF lastPos;

	CollisionChecker collisionCheckerAlgo;
	bool colliding;

	QWidget* status;
	ScoreUI* scoring;
	QWidget* stopWatch;	
};

OpenGLScene::OpenGLScene()
{
	d = new OpenGLSceneData;

	// Initialize few animation variables
	d->targetRotAngle = 0.0f;
	d->toggleTargetIncrDecrFlag = 0;
	d->projectileDisplaceX = 0.0f;
	d->projectileDisplaceY = 0.0f;
	d->fruitFall = -5.0f;
	d->colliding = false;
	d->toggleTargetHit = 0;
	
	
	// Create the widgets
	d->scoring = createScoringWidget();
	d->stopWatch = createStopWatchWidget();
	m_smileOrAngry = createImageWidget(":/Resources/Images/Arjunface_angry.jpg");

	// Load the background texture first
	m_fileName = ":/Resources/Images/Scenary.jpg";

	m_topScore = "";
	m_yourScore = "";

	// Widget settings for the draw context
	d->background.setTextureFileName(m_fileName);
	QWidget *widgets[] = { d->scoring, d->stopWatch, m_smileOrAngry };

	// Update the top score next
	updateUITopScore();

	for (uint i = 0; i < sizeof(widgets) / sizeof(*widgets); ++i) 
	{
        QGraphicsProxyWidget *proxy = new QGraphicsProxyWidget(0);
        proxy->setWidget(widgets[i]);
        addItem(proxy);
    }

	if( populateSceneAttribs() )
		setPreDefSceneAttribPositions();
		

	// Timer for 3D animations
	d->targetTimer = new QTimer(this);
	connect( d->targetTimer, SIGNAL(timeout()), this, SLOT(advanceTarget()) );
	d->targetTimer->start(30);

	d->projectileTimer = new QTimer(this);
	connect( d->projectileTimer, SIGNAL(timeout()), this, SLOT(advanceProjectile()) );
}

OpenGLScene::~OpenGLScene()
{
	delete d->status;
	delete d;
}

bool OpenGLScene::populateSceneAttribs()
{
	m_sceneItems = items();
	m_sceneAttribs["Score"] = m_sceneItems[2];
	m_sceneAttribs["Stopwatch"] = m_sceneItems[1];
	m_sceneAttribs["SmileOrAngry"] = m_sceneItems[0];

	return true;
}

void OpenGLScene::setPreDefSceneAttribPositions()
{
	if(m_sceneAttribs.contains("Score"))
	{
		QGraphicsItem* item = m_sceneAttribs.value("Score");
		item->setCacheMode(
            QGraphicsItem::DeviceCoordinateCache);
		const QRectF rect = item->boundingRect();
        item->setPos(0, 0);
	}

	if(m_sceneAttribs.contains("Stopwatch"))
	{
		QGraphicsItem* item = m_sceneAttribs.value("Stopwatch");
		item->setCacheMode(
            QGraphicsItem::DeviceCoordinateCache);
		const QRectF rect = item->boundingRect();
		item->setPos(300, 0);
	}

	if(m_sceneAttribs.contains("SmileOrAngry"))
	{
		QGraphicsItem* item = m_sceneAttribs.value("SmileOrAngry");
		item->setCacheMode(
            QGraphicsItem::DeviceCoordinateCache);
		const QRectF rect = item->boundingRect();
		item->setPos(550, 0);
	}
}

QDialog* OpenGLScene::createDialog(const QString& windowTitle) const
{
	QDialog *dialog = new QDialog(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint);

	dialog->setWindowOpacity(0.8);
	dialog->setWindowTitle(windowTitle);
	dialog->setLayout(new QVBoxLayout);

	return dialog;
}

QWidget* OpenGLScene::createWidget(const QString& windowTitle) const
{
	QWidget *widget = new QWidget(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint);

	widget->setWindowOpacity(0.8);
	widget->setWindowTitle(windowTitle);
	widget->setLayout(new QVBoxLayout);

	return widget;
}

ScoreUI* OpenGLScene::createScoringWidget()
{
	m_score = new ScoreUI;
	m_score->setWindowOpacity(0.8);
	return m_score;
}

QWidget* OpenGLScene::createStopWatchWidget()
{
	m_stopWatch = new Stopwatch;
	m_stopWatch->setWindowOpacity(0.8);
	
	connect(this, SIGNAL(collided()), m_stopWatch, SLOT(stop()));
	connect(this, SIGNAL(collided()), m_stopWatch, SLOT(updateTopScore()));
	connect(this, SIGNAL(collided()), this, SLOT(updateUITopScore()));

	return m_stopWatch;
}

QWidget* OpenGLScene::createImageWidget(const QString& filename)
{
	m_imageWidget = new ImageWidget;
	m_imageWidget->setFilename(filename);
	m_imageWidget->setWindowOpacity(0.8);
	return m_imageWidget;
}

void OpenGLScene::mousePressEvent(QGraphicsSceneMouseEvent* event)
{
	QGraphicsScene::mousePressEvent(event);
    if (event->isAccepted())
        return;

	if(event->button() == Qt::LeftButton)
		d->projectileTimer->start();

	if(event->button() == Qt::RightButton)
		d->lastPos = event->pos();

	event->accept();
	update();
}

void OpenGLScene::mouseMoveEvent(QGraphicsSceneMouseEvent* event)
{
	QGraphicsScene::mouseMoveEvent(event);
    if (event->isAccepted())
        return;

	//int dx = event->screenPos().x() - event->lastScreenPos().x();
	int dy = event->screenPos().y() - event->lastScreenPos().y();

	// Check if its a left move or right move accordingly
	// .. change the arrow coords. Do not do this movement
	// .. once the arrow is released
	if(dy < 0 && d->projectileDisplaceX < 5.0f)
		d->projectileDisplaceY += 0.05;
	if(dy > 0 && d->projectileDisplaceX < 5.0f)
		d->projectileDisplaceY -= 0.05;
	
	event->accept();
	update();
}

void OpenGLScene::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
	QGraphicsScene::mouseReleaseEvent(event);
    if (event->isAccepted())
        return;

	event->accept();
	update();
}

/* Toggle rotation angle increment/decrement for target
*/
void OpenGLScene::advanceTarget()
{
	if(d->toggleTargetIncrDecrFlag == 0)
	{
		d->targetRotAngle += 10;
		if(d->targetRotAngle == 90.0f)
			d->toggleTargetIncrDecrFlag = 1;
	}

	if(d->toggleTargetIncrDecrFlag == 1)
	{
		d->targetRotAngle -= 10;
		if(d->targetRotAngle == -90.0f)
			d->toggleTargetIncrDecrFlag = 0;
	}

	if(d->toggleTargetHit == 1)
		d->fruitFall-= 5.0f;

	update();
}

/* Projectile animation
*/
void OpenGLScene::advanceProjectile()
{
	d->projectileDisplaceX += 1.0;

	if(d->projectileDisplaceX > 25.0f)
	{
		d->projectileTimer->stop();
		d->projectileDisplaceX = 0.0f;
	}
	update();
}

void OpenGLScene::updateUITopScore()
{
	m_score->setTopScore( m_stopWatch->topScore() );
}

void OpenGLScene::drawBackground(QPainter* painter, const QRectF &)
{
	if (painter->paintEngine()->type() != QPaintEngine::OpenGL
		&& painter->paintEngine()->type() != QPaintEngine::OpenGL2)
    {
        qWarning("OpenGLScene: drawBackground needs a" 
			     "QOpenGLScene to be set as viewport on the graphics view");
        return;
    }

	painter->beginNativePainting();
	glClearColor(0.0f, 0.0f, 1.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Take care of the camera projections
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	//gluPerspective(70, width() / height(), 1.0, 15);
	glOrtho(-10.0, 10.0, -10.0, +10.0, -4.0, 15.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glEnable(GL_MULTISAMPLE);

	// MANDATORY !! BACKGROUND IMAGES NEED TO BE RENDERED FIRST
	glPushMatrix();
	GLfloat bckgOrigX = 10.0f;
	GLfloat bckgOrigY = 10.0f;
	GLfloat bckgOrigZ = 0.0f;
	glTranslatef(-bckgOrigX, -bckgOrigY, bckgOrigZ);
	glScalef( bckgOrigX * 2.0, bckgOrigY * 2.0, 1.0f);
	d->background.draw();
	glPopMatrix();

	// Draw the fruit target 
	glPushMatrix();
	glTranslatef(6.0f, 4.5f, 0.0f);
	glRotatef(d->targetRotAngle, 0.0f, 0.0f, 1.0f);
	glTranslatef(0.0f, -2.5f, 0.0f);
	d->target.draw();
	glPopMatrix();

	// The fruit
	glPushMatrix();
	glTranslatef(6.0f, 4.5f, 0.0f);
	glRotatef(d->targetRotAngle, 0.0f, 0.0f, 1.0f);
	glTranslatef(0.0f, d->fruitFall, 0.0f);
	GLfloat* pos = d->fruit.currentPosition();
	qDebug() << "Fruit = " << QString().setNum(pos[0]) 
			 << "," << QString().setNum(pos[1]) << "," << QString().setNum(pos[2]);
	GLfloat fruitDiffuse[] = { 0.7f, 0.0f, 0.0f, 1.0f };
	glMaterialfv(GL_FRONT, GL_DIFFUSE, fruitDiffuse);
	d->fruit.drawWithTheseProperties(0.5f, 15, 15, false);
	glPopMatrix();

	// Draw the arrow 
	glPushMatrix();
	glTranslatef(-8.0f, -5.0f, 0.0f);
	glRotatef(45, 0.0f, 0.0f, 1.0f);
	glTranslatef(d->projectileDisplaceX, d->projectileDisplaceY, 0.0f);
	d->arrow.draw();
	glPopMatrix();

	// Draw the arrow colliding object
	glPushMatrix();
	glTranslatef(-6.0f, -3.0f, 0.0f);
	glRotatef(45, 0.0f, 0.0f, 1.0f);
	glTranslatef(d->projectileDisplaceX, d->projectileDisplaceY, 0.0f);
	GLfloat* pos2 = d->arrowCollider.currentPosition();
	qDebug() << "Arrow = " << QString().setNum(pos2[0]) << "," << QString().setNum(pos2[1]) << "," << QString().setNum(pos2[2]);
	d->arrowCollider.drawWithTheseProperties(0.0, 0, 0, false);
	glPopMatrix();

	// Draw fruit pole without string next
	glPushMatrix();
	glTranslatef(6.0f, 2.0f, 0.0f);
	d->fruitPole.draw();
	glPopMatrix();

	if( d->collisionCheckerAlgo.colliding(pos, pos2) )
	{
		d->toggleTargetHit = 1;
		d->scoring->setFileName(":/Resources/Images/Score_3_2.jpg");
		m_imageWidget->setFilename(":/Resources/Images/Arjunface_happy.jpg");
		
		emit collided();
	}

	glDisable(GL_MULTISAMPLE);

	//----------
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	painter->beginNativePainting();
}